from functools import partial

import torch

from catalyst.utils.metrics import accuracy, multilabel_accuracy
from sklearn.metrics import f1_score, precision_score, recall_score
from torch import nn


class TorchMetric:

    def __init__(self, metric_fn, decode=True, metric_args=None):
        if metric_args is None:
            metric_args = {}
        self.activation = None
        self.decoder = None
        self.metric_fn = metric_fn
        self.metric_args = metric_args
        self._decode = decode
        self._is_binded = False

    def bind_to_task(self, task):
        self._is_binded = True
        self.activation = task.get_activation()
        self.decoder = task.get_decoder()

    def __call__(self, outputs, targets, activate=True):
        if not self._is_binded:
            raise ValueError(f'The metric is not binded to a task, but is already used.')
        outputs = torch.as_tensor(outputs)
        targets = torch.as_tensor(targets)

        if activate:
            outputs = self.activation(outputs)
        if self._decode:
            outputs = self.decoder(outputs)
        return self._invoke_metric(outputs, targets, self.metric_args)

    def _invoke_metric(self, outputs, targets, metric_args_dict):
        return self.metric_fn(outputs, targets, **metric_args_dict)


class BinaryAccuracy(TorchMetric):

    def __init__(self):
        super().__init__(accuracy, decode=True)

    def _invoke_metric(self, outputs, targets, metric_args_dict):
        if len(outputs.shape) <= 1:
            outputs = outputs.unsqueeze(dim=-1)
        return self.metric_fn(outputs, targets, **metric_args_dict)[0]


class ClassificationAccuracy(TorchMetric):

    def __init__(self, metric_args=None):
        if metric_args is None:
            metric_args = {'topk': [1]}
        super().__init__(accuracy, decode=False, metric_args=metric_args)

    def _invoke_metric(self, outputs, targets, metric_args_dict):
        topk = metric_args_dict['topk']
        results = self.metric_fn(outputs, targets, **metric_args_dict)
        dict_metrics = dict(zip(topk, results))
        return dict_metrics

    def empty_precondition_result(self):
        res = {}
        for metric_arg in self.metric_args['topk']:
            res[metric_arg] = torch.tensor(0.)
        return res


class NumpyMetric(TorchMetric):

    def __init__(self, metric_fn, decode=True):
        super().__init__(metric_fn, decode)

    def _invoke_metric(self, outputs, targets, metric_args_dict):
        if isinstance(outputs, torch.Tensor):
            outputs = outputs.detach().cpu().numpy()
        if isinstance(targets, torch.Tensor):
            targets = targets.cpu().numpy()
        return self.metric_fn(outputs, targets, **metric_args_dict)


class BinaryF1Score(NumpyMetric):

    def __init__(self):
        super().__init__(f1_score)


class BinaryPrecision(NumpyMetric):

    def __init__(self):
        super().__init__(precision_score)


class BinaryRecall(NumpyMetric):

    def __init__(self):
        super().__init__(recall_score)


class ClassificationNumpyMetric(NumpyMetric):

    def __init__(self, metric_fn, decode=True):
        super().__init__(metric_fn, decode=decode)

    def _invoke_metric(self, outputs, targets, metric_args_dict):
        if isinstance(outputs, torch.Tensor):
            outputs = outputs.detach().cpu().numpy()
        if isinstance(targets, torch.Tensor):
            targets = targets.cpu().numpy()
        outputs = outputs[..., 0]
        return self.metric_fn(outputs, targets, **metric_args_dict)


class ClassificationF1Score(ClassificationNumpyMetric):

    def __init__(self):
        super().__init__(partial(f1_score, average='micro'))


class ClassificationPrecision(ClassificationNumpyMetric):

    def __init__(self):
        super().__init__(partial(precision_score, average='micro'))


class ClassificationRecall(ClassificationNumpyMetric):

    def __init__(self):
        super().__init__(partial(recall_score, average='micro'))


class MeanAbsoluteError(TorchMetric):

    def __init__(self, decode=False):
        super().__init__(nn.L1Loss(), decode)


class MultiLabelClassificationAccuracy(TorchMetric):

    def __init__(self, metric_args=None):
        if metric_args is None:
            metric_args = {'threshold': 0.5}
        super().__init__(multilabel_accuracy, decode=True, metric_args=metric_args)

    def _invoke_metric(self, outputs, targets, metric_args_dict):
        # the threshold does not actually matter, since the outputs are already decoded, i.e they are already 1 and 0
        return self.metric_fn(outputs, targets, **metric_args_dict)


def get_default_binary_metrics():
    return (
             ('accuracy', BinaryAccuracy()),
             ('f1_score', BinaryF1Score()),
             ('precision', BinaryPrecision()),
             ('recall', BinaryRecall()),
    )


def get_default_bounded_regression_metrics():
    return (
        ('mean_absolute_error', MeanAbsoluteError()),
    )


def get_default_classification_metrics():
    return (
        ('accuracy', ClassificationAccuracy()),
        ('f1_score', ClassificationF1Score()),
        ('precision', ClassificationPrecision()),
        ('recall', ClassificationRecall()),
    )


def get_default_multilabel_classification_metrics():
    return (
        ('accuracy', MultiLabelClassificationAccuracy()),
    )



import torch

from torch.nn import functional as F


class DropoutMC:

    def __init__(self, num_samples=16, p=0.2):
        self.num_samples = num_samples
        self.p = p

    def create_samples(self, module, activation, *args, **kwargs):
        res = []
        for i in range(self.num_samples):
            new_args = [F.dropout(arg, p=self.p, training=True, inplace=False) for arg in args]
            new_kwargs = {key: F.dropout(value, p=self.p, training=True, inplace=False) for key, value in kwargs.items()}
            logits = module(*new_args, **new_kwargs)
            if activation is not None:
                logits = activation(logits)
            res.append(logits)
        return torch.stack(res, dim=0)





import torch

from pathlib import Path
from typing import Union, Iterable

from dnn_cool.converters import Values, Converters
from dnn_cool.runner import DnnCoolSupervisedRunner
from dnn_cool.task_flow import TaskFlow


def not_found_error_message(col, df):
    return f'Input column "{col}" not found in df. Dataframe has columns: {df.columns.tolist()}'


def assert_col_in_df(col, df):
    if isinstance(col, str):
        assert col in df, not_found_error_message(col, df)
    else:
        for col_s in col:
            assert col_s in df, not_found_error_message(col_s, df)


def create_values(df, output_col, converters):
    values_type = converters.type.guess(df, output_col)
    values = converters.values.to_values(df, output_col, values_type)
    return values


def create_leaf_task(df, col, converters):
    values = create_values(df, col, converters)
    task = converters.task.to_task(col, values.types[0], values.values[0])
    return task


def create_leaf_tasks(df, col, converters):
    if isinstance(col, str):
        return [create_leaf_task(df, col, converters)]

    res = []
    for col_s in col:
        res.append(create_leaf_task(df, col_s, converters))
    return res


def read_inputs(df, input_col, converters):
    if isinstance(input_col, str):
        values = create_values(df, input_col, converters)
        return values

    keys = []
    values = []
    types = []
    for col_s in input_col:
        vals = create_values(df, col_s, converters)
        keys.extend(vals.keys)
        values.extend(vals.values)
        types.extend(vals.types)

    return Values(keys=keys, values=values, types=types)


class UsedTasksTracer:

    def __init__(self):
        self.used_tasks = []

    def __getattr__(self, item):
        self.used_tasks.append(item)
        return self

    def __call__(self, *args, **kwargs):
        return self

    def __add__(self, other):
        return self

    def __invert__(self):
        return self

    def __or__(self, other):
        return self


class Project:

    def __init__(self, df,
                 input_col: Union[str, Iterable[str]],
                 output_col: Union[str, Iterable[str]],
                 project_dir: Union[str, Path],
                 converters: Converters):
        self.df = df
        assert_col_in_df(input_col, df)
        assert_col_in_df(output_col, df)

        self.project_dir = Path(project_dir)
        self.project_dir.mkdir(exist_ok=True)

        converters_file = self.project_dir / 'converters.pkl'
        if converters is None:
            converters = Converters()
        self.converters = converters
        if converters_file.exists():
            self.converters.load_state_dict(torch.load(converters_file))

        self.inputs = read_inputs(df, input_col, converters)
        self.leaf_tasks = create_leaf_tasks(df, output_col, converters)

        self.flow_tasks = []

        self._name_to_task = {}
        for leaf_task in self.leaf_tasks:
            self._name_to_task[leaf_task.get_name()] = leaf_task

        for i in range(len(self.inputs.keys)):
            self.converters.tensorboard_converters.col_to_type_mapping[self.inputs.keys[i]] = self.inputs.types[i]
        if not converters_file.exists():
            torch.save(self.converters.state_dict(), converters_file, pickle_protocol=4)

    def add_task_flow(self, task_flow: TaskFlow):
        self.flow_tasks.append(task_flow)
        self._name_to_task[task_flow.get_name()] = task_flow
        return task_flow

    def add_flow(self, func, flow_name=None):
        flow_name = func.__name__ if flow_name is None else flow_name
        flow = self.create_flow(func, flow_name)
        return self.add_task_flow(flow)

    def create_flow(self, flow_func, flow_name=None):
        flow_name = flow_func.__name__ if flow_name is None else flow_name
        used_tasks_tracer = UsedTasksTracer()
        flow_func(used_tasks_tracer, UsedTasksTracer(), UsedTasksTracer())
        used_tasks = []
        for used_task_name in used_tasks_tracer.used_tasks:
            task = self._name_to_task.get(used_task_name, None)
            if task is not None:
                used_tasks.append(task)
        return TaskFlow(flow_name, used_tasks, flow_func, self.inputs)

    def get_all_tasks(self):
        return self.leaf_tasks + self.flow_tasks

    def get_full_flow(self) -> TaskFlow:
        return self.flow_tasks[-1]

    def get_task(self, task_name):
        return self._name_to_task[task_name]

    def runner(self, model,
               early_stop=True,
               balance_dataparallel_memory=False,
               runner_name=None,
               train_test_val_indices=None):
        return DnnCoolSupervisedRunner(self, model,
                                       early_stop=early_stop,
                                       balance_dataparallel_memory=balance_dataparallel_memory,
                                       runner_name=runner_name,
                                       train_test_val_indices=train_test_val_indices)



import numpy as np
import torch
from sklearn.model_selection import train_test_split
from torch.utils.data import Dataset, Subset
from torchvision import transforms


def any_value(outputs):
    for key, value in outputs.items():
        if not key.startswith('precondition') and not key == 'gt':
            return value


def torch_split_dataset(dataset, test_size=0.2, random_state=None):
    train_indices, test_indices = split_dataset(dataset, test_size, random_state=random_state)
    train_dataset = Subset(dataset, train_indices)
    test_dataset = Subset(dataset, test_indices)
    return train_dataset, test_dataset


def split_dataset(dataset, test_size=0.2, random_state=None):
    X, y = np.arange(len(dataset)), np.arange(len(dataset))
    train_indices, test_indices, _, _ = train_test_split(X, y, test_size=test_size, random_state=random_state)
    return train_indices, test_indices


def train_test_val_split(df, random_state=None):
    train_indices, val_indices = split_dataset(df, test_size=0.2, random_state=random_state)
    rel_test_indices, rel_val_indices = split_dataset(val_indices, test_size=0.5, random_state=random_state)
    return train_indices, val_indices[rel_test_indices], val_indices[rel_val_indices]


class TransformedSubset(Dataset):
    r"""
    Subset of a dataset at specified indices, with additionally applied transforms.

    Arguments:
        dataset (Dataset): The whole Dataset
        indices (sequence): Indices in the whole set selected for subset
    """
    def __init__(self, dataset, indices, sample_transforms=None):
        self.dataset = dataset
        self.indices = indices
        self.transforms = sample_transforms

    def __getitem__(self, idx):
        r = self.dataset[self.indices[idx]]
        if self.transforms is None:
            return r
        return self.transforms(r)

    def __len__(self):
        return len(self.indices)


def to_broadcastable_shape(tensor1, tensor2):
    if len(tensor1.shape) == len(tensor2.shape):
        return tensor1, tensor2
    if len(tensor2.shape) > len(tensor1.shape):
        return to_broadcastable_smaller(tensor1, tensor2)
    tensor2, tensor1 = to_broadcastable_smaller(tensor2, tensor1)
    return tensor1, tensor2


def to_broadcastable_smaller(tensor1, tensor2):
    assert tensor1.shape == tensor2.shape[:len(tensor1.shape)]
    old_axes = tensor1.shape
    new_axes = [1] * (len(tensor2.shape) - len(tensor1.shape))
    return tensor1.view(*old_axes, *new_axes), tensor2


def check_if_should_divide(sample):
    if sample.dtype == np.uint8:
        return True
    if sample.dtype == torch.uint8:
        return True
    if sample.max() > 1.:
        return True
    return False


def check_if_should_transpose(sample):
    return sample.shape[-1] == 3


class ImageNetNormalizer:

    def __init__(self, should_divide_by_255=None, should_transpose=None):
        self.should_divide_by_255 = should_divide_by_255
        self.should_transpose = should_transpose
        self.normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                              std=[0.229, 0.224, 0.225])

    def __call__(self, sample):
        if self.should_divide_by_255 is None:
            self.should_divide_by_255 = check_if_should_divide(sample)
        if self.should_transpose is None:
            self.should_transpose = check_if_should_transpose(sample)
        if self.should_divide_by_255:
            sample = sample / 255.
        X = torch.tensor(sample).float()
        if self.should_transpose:
            X = X.permute(2, 0, 1)
        X = self.normalize(X)
        return X



from typing import Dict

from dataclasses import dataclass, field

from dnn_cool.visitors import LeafVisitor, VisitorOut, RootCompositeVisitor


class FilterVisitor(LeafVisitor):

    def __init__(self, task, prefix):
        super().__init__(task, prefix)

    def full_result(self, preds, targets):
        return FilterParams(predictions={self.path: preds}, targets={self.path: targets})

    def empty_result(self):
        return FilterParams(predictions={}, targets={})

    def preconditioned_result(self, preds, targets):
        return FilterParams(predictions={self.path: preds}, targets={self.path: targets})


@dataclass
class FilterParams(VisitorOut):
    predictions: Dict = field(default_factory=lambda: {})
    targets: Dict = field(default_factory=lambda: {})

    def __iadd__(self, other):
        self.predictions.update(other.predictions)
        self.targets.update(other.targets)
        return self

    def reduce(self):
        return {
            'predictions': self.predictions,
            'targets': self.targets
        }


class FilterCompositeVisitor(RootCompositeVisitor):

    def __init__(self, task_flow, prefix):
        super().__init__(task_flow, FilterVisitor, FilterParams, prefix=prefix)

    def load_tuned(self, tuned_params):
        tasks = self.task_flow.get_all_children()
        for path, task in tasks.items():
            task.get_decoder().load_tuned(tuned_params[path])



from typing import Dict, Tuple

import torch

from dataclasses import dataclass
from torch.utils.data.dataset import Dataset

from dnn_cool.dsl import IFeaturesDict, IOut, ICondition, IFlowTask


class LeafTaskDataset(Dataset):

    def __init__(self, inputs, labels):
        self.inputs = inputs
        self.labels = labels
        assert len(self.inputs) == len(self.labels), f'The provided inputs and labels are of different length: ' \
                                                     f'{len(self.inputs)} and {len(self.labels)}'

    def __getitem__(self, item):
        return self.inputs[item], self.labels[item]

    def __len__(self):
        return len(self.inputs)


def discover_index_holder(*args, **kwargs):
    all_args = [*args, *kwargs.values()]

    for arg in all_args:
        if isinstance(arg, IndexHolder):
            return arg


class FlowDatasetDecorator:

    def __init__(self, task, prefix, labels):
        self.task_name = task.get_name()
        self.available = task.get_available_func()(labels)
        self.prefix = prefix
        self.arr = labels

    def __call__(self, *args, **kwargs):
        index_holder = discover_index_holder(*args, **kwargs)
        key = self.prefix + self.task_name
        data = {
            key: self.arr[index_holder.item]
        }
        available = {
            key: self.available[index_holder.item]
        }
        return FlowDatasetDict(self.prefix, data, available)


class IndexHolder(IFeaturesDict):

    def __init__(self, item):
        self.item = item

    def __getattr__(self, item):
        return self


@dataclass
class FlowDatasetPrecondition(ICondition):
    prefix: str
    path: str
    precondition: torch.Tensor

    def __invert__(self):
        return self

    def __and__(self, other):
        return self


class TensorDict(dict):

    def __len__(self):
        for key, value in self.items():
            if key != 'gt':
                return len(value)
        return super().__len__()


class FlowDatasetDict(IOut):

    def __init__(self, prefix, data=None, available=None):
        self.prefix = prefix
        self.data = data if data is not None else {}
        self.available = available if available is not None else {}
        self.gt = {}

    def __iadd__(self, other):
        for key, value in other.data.items():
            self.data[key] = value
        self.gt.update(other.gt)
        self.available.update(other.available)
        return self

    def __getattr__(self, item):
        return FlowDatasetPrecondition(prefix=self.prefix,
                                       path=self.prefix + item,
                                       precondition=self.data[self.prefix + item])

    def __or__(self, other: FlowDatasetPrecondition):
        self.gt.update({other.path: other.precondition.bool()})
        return self

    def to_dict(self, X):
        y = {}
        for key, value in self.data.items():
            targets = value
            y[key] = targets
        X['gt'] = self.gt
        X['gt']['_availability'] = self.available
        X['gt']['_targets'] = y
        return TensorDict(X), y


class FlowDataset(Dataset, IFlowTask):

    def __init__(self, task_flow, prefix=''):
        """
        Creates a dataset object for a given task flow.

        :param task_flow: The task flow object.

        :param prefix: The prefix path (will be appended before the task flow name)

        """
        self._task_flow = task_flow
        # Save a reference to the flow function of the original class
        # We will then call it by replacing the self, this way effectively running
        # it with this class. And this class stores Pytorch modules as class attributes
        self.flow = task_flow.get_flow_func()

        self.n = None
        for key, task in task_flow.tasks.items():
            if not task.has_children():
                labels_instance = FlowDatasetDecorator(task, prefix, task.get_labels())
                self.n = len(labels_instance.arr)
                setattr(self, key, labels_instance)
            else:
                instance = FlowDataset(task, prefix=f'{prefix}{task.get_name()}.')
                setattr(self, key, instance)
        self.prefix = prefix

    def __getitem__(self, item: int) -> Tuple[Dict, Dict]:
        """
        This method is going to execute the flow given in the constructor, passing around the index variable `item` in
        the class :class:`IndexHolder`.

        :param item: The index in the dataset.

        :return: A tuple X, y of two dictionaries
        """
        flow_dataset_dict = self.flow(self, IndexHolder(item), FlowDatasetDict(self.prefix, {}))
        inputs = self._task_flow.get_inputs()
        if inputs is None:
            raise ValueError(f'Cannot build a dataset, since the inputs are not provided. You have to provide them'
                             f' in the constructor of the TaskFlow class.')
        X = inputs[item]
        # X has to be a dict, because we have to attach gt.
        if not isinstance(X, dict):
            X = {
                'inputs': X
            }
        return flow_dataset_dict.to_dict(X)

    def __len__(self):
        return self.n

    def __call__(self, *args, **kwargs):
        index_holder = discover_index_holder(*args, **kwargs)
        return self.flow(self, index_holder, FlowDatasetDict(self.prefix))



from typing import Dict

from dataclasses import dataclass, field

from dnn_cool.visitors import LeafVisitor, VisitorOut, RootCompositeVisitor


class TuningVisitor(LeafVisitor):

    def __init__(self, task, prefix):
        super().__init__(task, prefix)

    def full_result(self, preds, targets):
        return TunedParams(self.decoder.tune(preds, targets))

    def empty_result(self):
        return TunedParams({self.path: {}})

    def preconditioned_result(self, preds, targets):
        tuned_params = self.decoder.tune(preds, targets)
        return TunedParams({self.path: tuned_params})


@dataclass
class TunedParams(VisitorOut):
    data: Dict = field(default_factory=lambda: {})

    def __iadd__(self, other):
        self.data.update(other.data)
        return self

    def reduce(self):
        return self.data


class TunerVisitor(RootCompositeVisitor):

    def __init__(self, task_flow, prefix):
        super().__init__(task_flow, TuningVisitor, TunedParams, prefix=prefix)

    def load_tuned(self, tuned_params):
        tasks = self.task_flow.get_all_children()
        for path, task in tasks.items():
            task.get_decoder().load_tuned(tuned_params[path])



from collections import OrderedDict
from typing import Iterator, Mapping
from dataclasses import dataclass, field
from functools import partial
from pathlib import Path
from time import time
from typing import Dict, Tuple, Callable, Optional, Any, Union, List
from torch import nn

import numpy as np
import torch
from catalyst.dl import SupervisedRunner, EarlyStoppingCallback, InferCallback, State, Callback
from catalyst.utils import load_checkpoint, unpack_checkpoint, any2device
from torch import optim
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader, Dataset

from dnn_cool.catalyst_utils import InterpretationCallback, TensorboardConverters, ReplaceGatherCallback
from dnn_cool.utils import TransformedSubset, train_test_val_split


@dataclass
class TrainingArguments(Mapping):
    num_epochs: int
    criterion: Optional[Callable] = None
    model: Optional[Callable] = None
    optimizer: Optional[Optimizer] = None
    scheduler: Optional[Any] = None
    logdir: Optional[Union[str, Path]] = None
    loaders: Optional[Dict[str, DataLoader]] = None
    callbacks: Optional[Union[List[Callback], Dict[str, Callback]]] = None
    fp16: Union[Dict, bool] = None
    catalyst_args: Dict = field(default_factory=lambda: {})
    train_transforms: Callable = None
    val_transforms: Callable = None

    def __getitem__(self, k):
        if hasattr(self, k):
            return getattr(self, k)
        return self.catalyst_args[k]

    def __len__(self) -> int:
        return len(self.__dataclass_fields__) + len(self.catalyst_args)

    def __iter__(self) -> Iterator:
        res = {}
        for field_name in self.__dataclass_fields__:
            if field_name != 'catalyst_args':
                attr = getattr(self, field_name)
                if attr is not None:
                    res[field_name] = attr
        for key, value in self.catalyst_args.items():
            res[key] = value
        return iter(res)


class InferDictCallback(InferCallback):

    def __init__(self, out_key='logits', loaders_to_skip=(), *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.loaders_to_skip = loaders_to_skip
        self.out_key = out_key
        self.predictions = {}
        self.targets = {}
        self.__current_store = None

    def on_loader_start(self, state: State):
        self.predictions[state.loader_key] = {}
        self.targets[state.loader_key] = {}

    def on_batch_end(self, state: State):
        dct = state.output[self.out_key]
        if '_device|overall|_n' in dct:
            if self.__current_store is not None:
                self.__current_store(loader_name=state.loader_key)
            return
        dct = {key: value.detach().cpu().numpy() for key, value in dct.items() if key != 'gt'}
        loader_name = state.loader_key
        targets = state.input['targets']
        self.update_storage(loader_name, dct, targets)

    def on_dataparallel_gather(self, dct):
        self.__current_store = partial(self.update_storage, dct=dct, targets=dct['gt']['_targets'])

    def update_storage(self, loader_name, dct, targets):
        for key, value in dct.items():
            if key == 'gt':
                continue
            if key not in self.predictions[loader_name]:
                self.predictions[loader_name][key] = []
            if isinstance(value, list):
                self.predictions[loader_name][key].extend(value)
            else:
                self.predictions[loader_name][key].append(value)
        for key, value in targets.items():
            if key == 'gt':
                continue
            if key not in self.targets[loader_name]:
                self.targets[loader_name][key] = []
            if isinstance(value, list):
                self.targets[loader_name][key].extend(value)
            else:
                self.targets[loader_name][key].append(value.detach().cpu().numpy())

    def on_loader_end(self, state: State):
        self.predictions[state.loader_key] = {
            key: np.concatenate(value, axis=0)
            for key, value in self.predictions[state.loader_key].items()
        }

        self.targets[state.loader_key] = {
            key: np.concatenate(value, axis=0)
            for key, value in self.targets[state.loader_key].items()
        }


class DnnCoolSupervisedRunner(SupervisedRunner):

    def __init__(self, project, model,
                 early_stop: bool = True,
                 balance_dataparallel_memory: bool = False,
                 runner_name=None,
                 train_test_val_indices=None):
        self.task_flow = project.get_full_flow()

        self.default_criterion = self.task_flow.get_loss()
        self.balance_dataparallel_memory = balance_dataparallel_memory

        self.default_callbacks = []
        if self.balance_dataparallel_memory:
            self.default_callbacks.append(ReplaceGatherCallback(self.task_flow))
        self.default_callbacks.extend(self.default_criterion.catalyst_callbacks())
        self.default_optimizer = partial(optim.AdamW, lr=1e-4)
        self.default_scheduler = ReduceLROnPlateau
        self.project_dir: Path = project.project_dir
        self.project_dir.mkdir(exist_ok=True)
        runner_name = f'{self.task_flow.get_name()}_{time()}' if runner_name is None else runner_name
        self.default_logdir = f'./logdir_{runner_name}'

        if early_stop:
            self.default_callbacks.append(EarlyStoppingCallback(patience=5))

        (self.project_dir / self.default_logdir).mkdir(exist_ok=True)
        if train_test_val_indices is None:
            train_test_val_indices = project_split(project.df, self.project_dir / self.default_logdir)
        else:
            save_split(self.project_dir / self.default_logdir, train_test_val_indices)
        self.train_test_val_indices = train_test_val_indices
        self.tensor_loggers = project.converters.tensorboard_converters
        super().__init__(model=model)

    def train(self, *args, **kwargs):
        kwargs['criterion'] = kwargs.get('criterion', self.default_criterion)
        kwargs['model'] = kwargs.get('model', self.model)

        if 'optimizer' not in kwargs:
            model = kwargs['model']
            optimizable_params = filter(lambda p: p.requires_grad, model.parameters())
            kwargs['optimizer'] = self.default_optimizer(params=optimizable_params)

        if 'scheduler' not in kwargs:
            kwargs['scheduler'] = self.default_scheduler(kwargs['optimizer'])

        if 'logdir' not in kwargs:
            kwargs['logdir'] = self.default_logdir
        kwargs['logdir'] = self.project_dir / kwargs['logdir']
        kwargs['num_epochs'] = kwargs.get('num_epochs', 50)

        if 'loaders' not in kwargs:
            datasets, kwargs['loaders'] = self.get_default_loaders()

        default_callbacks = [self.create_interpretation_callback(**kwargs)] + self.default_callbacks
        kwargs['callbacks'] = kwargs.get('callbacks', default_callbacks)
        super().train(*args, **kwargs)

    def infer(self, *args, **kwargs):
        default_datasets, default_loaders = self.get_default_loaders(shuffle_train=False)
        kwargs['loaders'] = kwargs.get('loaders', default_loaders)
        kwargs['datasets'] = kwargs.get('datasets', default_datasets)

        logdir = self.project_dir / Path(kwargs.get('logdir', self.default_logdir))
        kwargs['logdir'] = logdir
        interpretation_callback = self.create_interpretation_callback(**kwargs)
        infer_dict_callback = InferDictCallback()
        default_callbacks = OrderedDict([("interpretation", interpretation_callback),
                                         ("inference", infer_dict_callback)])
        if self.balance_dataparallel_memory:
            replace_gather_callback = ReplaceGatherCallback(self.task_flow, infer_dict_callback)
            default_callbacks["dataparallel_reducer"] = replace_gather_callback
        kwargs['callbacks'] = kwargs.get('callbacks', default_callbacks)
        kwargs['model'] = kwargs.get('model', self.model)
        store = kwargs.pop('store', True)
        kwargs.pop('loader_names_to_skip_in_interpretation', ())
        del kwargs['datasets']
        super().infer(*args, **kwargs)
        res = {}
        if 'inference' in kwargs['callbacks']:
            res['logits'] = kwargs['callbacks']['inference'].predictions
            res['targets'] = kwargs['callbacks']['inference'].targets
        if 'interpretation' in kwargs['callbacks']:
            res['interpretations'] = kwargs['callbacks']['interpretation'].interpretations

        if store:
            out_dir = logdir / 'infer'
            out_dir.mkdir(exist_ok=True)
            for key in res:
                torch.save(res[key], out_dir / f'{key}.pkl', pickle_protocol=4)
        return res

    def create_interpretation_callback(self, **kwargs) -> InterpretationCallback:
        tensorboard_converters = TensorboardConverters(
            logdir=kwargs['logdir'],
            tensorboard_loggers=self.tensor_loggers,
            datasets=kwargs.get('datasets', self.get_default_datasets(**kwargs))
        )
        loaders_to_skip = kwargs.get('loader_names_to_skip_in_interpretation', ())
        interpretation_callback = InterpretationCallback(self.task_flow, tensorboard_converters, loaders_to_skip)
        return interpretation_callback

    def get_default_loaders(self, shuffle_train=True,
                            collator=None,
                            batch_size_per_gpu=32) -> Tuple[Dict[str, Dataset], Dict[str, DataLoader]]:
        datasets = self.get_default_datasets()
        train_dataset = datasets['train']
        val_dataset = datasets['valid']
        test_dataset = datasets['test']
        bs = max(batch_size_per_gpu, batch_size_per_gpu * torch.cuda.device_count())
        train_loader = DataLoader(train_dataset, batch_size=bs, shuffle=shuffle_train, collate_fn=collator)
        val_loader = DataLoader(val_dataset, batch_size=bs, shuffle=False, collate_fn=collator)
        test_loader = DataLoader(test_dataset, batch_size=batch_size_per_gpu * torch.cuda.device_count(), shuffle=False,
                                 collate_fn=collator)
        loaders = OrderedDict({
            'train': train_loader,
            'valid': val_loader,
        })

        # Rename 'train' loader and dataset, since catalyst does not allow inference on train dataset.
        if not shuffle_train:
            loaders['infer'] = loaders['train']
            del loaders['train']
            datasets['infer'] = datasets['train']
            del datasets['train']
            loaders['test'] = test_loader
        return datasets, loaders

    def get_default_datasets(self, **kwargs) -> Dict[str, Dataset]:
        dataset = self.task_flow.get_dataset()
        if self.train_test_val_indices is None:
            raise ValueError(f'You must supply either a `loaders` parameter, or give `train_test_val_indices` via'
                             f'constructor.')
        train_indices, test_indices, val_indices = self.train_test_val_indices
        train_dataset = TransformedSubset(dataset, train_indices, sample_transforms=kwargs.get('train_transforms'))
        val_dataset = TransformedSubset(dataset, val_indices, sample_transforms=kwargs.get('val_transforms'))
        test_dataset = TransformedSubset(dataset, test_indices, sample_transforms=kwargs.get('val_transforms'))

        datasets = {
            'train': train_dataset,
            'valid': val_dataset,
            'test': test_dataset,
        }

        datasets['infer'] = datasets[kwargs.get('target_loader', 'valid')]
        return datasets

    def batch_to_device(self, batch, device) -> Mapping[str, torch.Tensor]:
        return any2device(batch, device)

    def batch_to_model_device(self, batch) -> Mapping[str, torch.Tensor]:
        return any2device(batch, next(self.model.parameters()).device)

    def best(self) -> nn.Module:
        model = self.model
        checkpoint_path = str(self.project_dir / self.default_logdir / 'checkpoints' / 'best_full.pth')
        ckpt = load_checkpoint(checkpoint_path)
        unpack_checkpoint(ckpt, model)

        thresholds_path = self.project_dir / self.default_logdir / 'tuned_params.pkl'
        if not thresholds_path.exists():
            return model
        tuned_params = torch.load(thresholds_path)
        self.task_flow.get_decoder().load_tuned(tuned_params)
        return model

    def tune(self, loader_name='valid', store=True) -> Dict:
        res = self.load_inference_results()
        decoder = self.task_flow.get_decoder()
        tuned_params = decoder.tune(res['logits'][loader_name], res['targets'][loader_name])
        if store:
            out_path = self.project_dir / self.default_logdir / 'tuned_params.pkl'
            torch.save(tuned_params, out_path)
        return tuned_params

    def load_inference_results(self) -> Dict:
        logdir = self.project_dir / self.default_logdir
        out_dir = logdir / 'infer'
        out_dir.mkdir(exist_ok=True)
        res = {}
        for key in ['logits', 'targets', 'interpretations']:
            file_path = out_dir / f'{key}.pkl'
            if file_path.exists():
                res[key] = torch.load(out_dir / f'{key}.pkl')
        return res

    def load_tuned(self) -> Dict:
        tuned_params = torch.load(self.project_dir / self.default_logdir / 'tuned_params.pkl')
        self.task_flow.get_decoder().load_tuned(tuned_params)
        return tuned_params

    def evaluate(self, loader_name='test'):
        res = self.load_inference_results()
        self.load_tuned()
        evaluator = self.task_flow.get_evaluator()
        df = evaluator(res['logits'][loader_name], res['targets'][loader_name])
        df.to_csv(self.project_dir / self.default_logdir / 'evaluation.csv', index=False)
        return df


def split_already_done(df, project_dir):
    total_len = 0
    for i, split_name in enumerate(['train', 'test', 'val']):
        split_path = project_dir / f'{split_name}_indices.npy'
        if not split_path.exists():
            return False
        total_len += len(np.load(split_path))

    if total_len != len(df):
        return False
    return True


def read_split(project_dir):
    res = []
    for i, split_name in enumerate(['train', 'test', 'val']):
        split_path = project_dir / f'{split_name}_indices.npy'
        res.append(np.load(split_path))
    return res


def save_split(project_dir, res):
    for i, split_name in enumerate(['train', 'test', 'val']):
        split_path = project_dir / f'{split_name}_indices.npy'
        np.save(split_path, res[i])


def project_split(df, project_dir):
    if split_already_done(df, project_dir):
        return read_split(project_dir)
    res = train_test_val_split(df)
    save_split(project_dir, res)
    return res



import torch

from torch import nn


def dummy_tensor(inputs):
    return nn.Parameter(data=torch.ones_like(inputs))


def dummy_dict(inputs):
    res = nn.ParameterDict()
    for key, value in inputs.items():
        if key != 'gt' and not key.startswith('precondition'):
            res[key] = dummy_tensor(value)
    return res


class RFModel(nn.Module):

    def __init__(self, inputs, model):
        super().__init__()
        self.dummy = dummy_dict(inputs) if isinstance(inputs, dict) else dummy_tensor(inputs)
        self.model = model

    def forward(self, x):
        if isinstance(x, torch.Tensor):
            return self.model(self.dummy * x)
        for key, value in self.dummy.items():
            x[key] *= value
        return self.model(x)


def compute_receptive_field(model, inputs, out_mask, marker=5e8):
    rf_model = RFModel(inputs, model)
    pred = rf_model(inputs)
    target = mark_predictions(pred, out_mask, marker)
    rf_backprop(pred, target)
    rf_mask = get_rf_mask(rf_model)
    return rf_mask


def get_rf_mask(rf_model, eps=1e-4):
    if isinstance(rf_model.dummy, torch.Tensor):
        return abs(rf_model.dummy.grad) > eps
    rf_mask = {}
    for key, value in rf_model.dummy.items():
        rf_mask[key] = abs(value.grad) > eps
    return rf_mask


def rf_backprop(pred, target):
    if isinstance(pred, torch.Tensor):
        mse = nn.MSELoss()
        loss = mse(pred, target)
        loss.backward()
        return
    mse = nn.MSELoss()
    losses = []
    for key, value in pred.items():
        losses.append(mse(value, target[key]))
    loss = torch.stack(losses).mean()
    loss.backward()


def mark_predictions(pred, out_mask, marker):
    if isinstance(pred, torch.Tensor):
        target = pred.detach().clone()
        target[out_mask] += marker
        return target
    assert isinstance(pred, dict)
    target = {}
    for key, value in pred.items():
        target[key] = value.detach().clone()
        target[key][out_mask[key]] += marker
    return target



from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
import torch
import matplotlib.image as mpimg

from sklearn.preprocessing import MultiLabelBinarizer


def binary_value_converter(values):
    values[np.isnan(values.astype(float))] = -1
    return torch.tensor(values.astype(float)).float().unsqueeze(dim=-1)


def classification_converter(values):
    values[np.isnan(values)] = -1
    return torch.tensor(values).long()


class MultiLabelValuesConverter:

    def __init__(self):
        self.binarizer = MultiLabelBinarizer()
        self.is_fit = False

    def __call__(self, values):
        available_labels = values[~pd.isna(values)]
        labels = []
        for i in range(len(available_labels)):
            split = available_labels.iloc[i].split(',')
            labels.append(list(filter(lambda x: len(x) > 0, split)))
        if not self.is_fit:
            one_hot_labels = self.binarizer.fit_transform(labels)
            self.is_fit = True
        else:
            one_hot_labels = self.binarizer.transform(labels)
        res = np.ones((len(values), one_hot_labels.shape[1]), dtype=np.float32) * -1.
        res[~pd.isna(values)] = one_hot_labels
        return res

    def state_dict(self):
        return {
            'binarizer': self.binarizer,
            'is_fit': self.is_fit
        }

    def load_state_dict(self, state_dict):
        self.binarizer = state_dict['binarizer']
        self.is_fit = state_dict['is_fit']


class ImageCoordinatesValuesConverter:

    def __init__(self, dim: Union[float, np.ndarray]):
        self.dim = dim

    def __call__(self, values):
        values = values.astype(float) / self.dim
        values[np.isnan(values)] = -1
        return torch.tensor(values).float().unsqueeze(dim=-1)


class ImagesFromFileSystem:

    def __init__(self, base_path, values):
        self.base_path = base_path
        self.values = values

    def __getitem__(self, item):
        return mpimg.imread(self.base_path / self.values[item])

    def __len__(self):
        return len(self.values)


class ImagesFromFileSystemConverter:

    def __init__(self, base_path):
        self.base_path = Path(base_path)

    def __call__(self, values):
        return ImagesFromFileSystem(self.base_path, values)



from collections import OrderedDict
from functools import partial
from pathlib import Path

import cv2
import numpy as np
import pandas as pd
import torch
from torch import nn
from torch.utils.data import DataLoader

from dnn_cool.converters import TypeGuesser, ValuesConverter, TaskConverter, Converters
from dnn_cool.decoders import BoundedRegressionDecoder
from dnn_cool.project import Project
from dnn_cool.task_converters import To
from dnn_cool.task_flow import BoundedRegressionTask, BinaryClassificationTask, ClassificationTask, \
    MultilabelClassificationTask
from dnn_cool.utils import torch_split_dataset
from dnn_cool.value_converters import binary_value_converter, classification_converter, \
    ImageCoordinatesValuesConverter, MultiLabelValuesConverter


def add_random_noise(img):
    n = 40
    i = np.random.randint(0, 64, size=n)
    j = np.random.randint(0, 64, size=n)
    c = np.random.randint(0, 3, size=n)
    values = np.random.randint(0, 255, size=n)
    img[i, j, c] = values


def generate_camera_blocked_image():
    img = np.zeros((64, 64, 3), dtype=np.uint8)
    # randomly add some noise
    add_random_noise(img)
    return img, {'camera_blocked': True}


def generate_door_open_image():
    img = np.ones((64, 64, 3), dtype=np.uint8) * 255
    # randomly add some noise
    add_random_noise(img)
    return img, {'camera_blocked': False, 'door_open': True, 'person_present': False}


def generate_door_closed_image(door_locked):
    img, res = generate_door_open_image()

    offsets = np.random.randint(-10, 10, size=4)
    x1, y1 = int(10 + offsets[0]), int(10 + offsets[1])
    x2, y2 = int(50 + offsets[2]), int(50 + offsets[2])

    img = cv2.rectangle(img, (x1, y1), (x2, y2), color=(139, 69, 19), thickness=-1)

    if door_locked:
        lock_start = x1, int((y1 + y2) / 2)
        lock_end = lock_start[0] + 10, lock_start[1] + 10
        img = cv2.rectangle(img, lock_start, lock_end, color=(0, 0, 255), thickness=-1)

    res['door_locked'] = door_locked
    res['camera_blocked'] = False
    res['door_open'] = False
    return img, res


def select_facial_characteristics():
    choices = [
        [(255, 0, 255), (0, 2)],
        [(0, 255, 255), (1, 2)],
        [(255, 255, 0), (0, 1)],
        [(0, 0, 0), ()],
        [(255, 0, 0), (0,)],
        [(0, 255, 0), (1,)],
        [(0, 0, 255), (2,)],
    ]
    choice = np.random.randint(0, len(choices), size=1)[0]
    return choices[choice]


def draw_person(img, res, shirt_type='blue'):
    head_radius = 6
    offsets = np.random.randint(-4, 4, size=4)
    head = int(30 + offsets[0]), int(10 + offsets[1])

    color, face_characteristics = select_facial_characteristics()
    img = cv2.circle(img, head, head_radius, color=color, thickness=-1)

    res['person_present'] = True
    res['face_x1'] = head[0] - head_radius
    res['face_y1'] = head[1] - head_radius
    res['face_w'] = 2 * head_radius
    res['face_h'] = 2 * head_radius
    res['facial_characteristics'] = ','.join(map(str, face_characteristics))

    offsets = np.random.randint(-2, 2, size=4)
    d = head_radius * 2
    rec_start = head[0] - head_radius + offsets[0], head[1] + head_radius + offsets[1]
    rec_end = rec_start[0] + d + offsets[2], rec_start[1] + 30 + offsets[3]

    if shirt_type == 'blue':
        color = (0, 0, 255)
        shirt_label = 0
    elif shirt_type == 'red':
        color = (255, 0, 0)
        shirt_label = 1
    elif shirt_type == 'yellow':
        color = (255, 255, 0)
        shirt_label = 2
    elif shirt_type == 'cyan':
        color = (0, 255, 255)
        shirt_label = 3
    elif shirt_type == 'magenta':
        color = (255, 0, 255)
        shirt_label = 4
    elif shirt_type == 'green':
        color = (0, 255, 0)
        shirt_label = 5
    else:
        # black
        color = (0, 0, 0)
        shirt_label = 6

    cv2.rectangle(img, rec_start, rec_end, color=color, thickness=-1)

    res['body_x1'] = rec_start[0]
    res['body_y1'] = rec_start[1]
    res['body_w'] = rec_end[0] - rec_start[0]
    res['body_h'] = rec_end[1] - rec_start[1]
    res['shirt_type'] = shirt_label

    return img, res


def generate_image_with_person(shirt_type='blue'):
    img, res = generate_door_open_image()
    img, res = draw_person(img, res, shirt_type)
    return img, res


def generate_sample():
    generators = [generate_camera_blocked_image,
                  generate_door_open_image,
                  partial(generate_door_closed_image, door_locked=True),
                  partial(generate_door_closed_image, door_locked=False),
                  partial(generate_image_with_person, shirt_type='blue'),
                  partial(generate_image_with_person, shirt_type='red'),
                  partial(generate_image_with_person, shirt_type='yellow'),
                  partial(generate_image_with_person, shirt_type='cyan'),
                  partial(generate_image_with_person, shirt_type='magenta'),
                  partial(generate_image_with_person, shirt_type='green'),
                  partial(generate_image_with_person, shirt_type='black')
                  ]
    choice = np.random.randint(0, len(generators), size=1)[0]
    return generators[choice]()


def create_df_and_images_tensor(n=int(1e4), cache_file=Path('dnn_cool_synthetic_dataset.pkl')):
    if cache_file.exists():
        return torch.load(cache_file)
    imgs = []
    rows = []
    names = []
    for i in range(n):
        img, row = generate_sample()
        imgs.append(torch.tensor(img).permute(2, 0, 1))
        rows.append(row)
        names.append(f'{i}.jpg')

    df = pd.DataFrame(rows)
    df['syn_img'] = names
    df.loc[:5, 'camera_blocked'] = np.nan
    res = torch.stack(imgs, dim=0).float() / 255., df
    torch.save(res, cache_file)
    return res


def synthetic_dataset_preparation(n=int(1e4)):
    imgs, df = create_df_and_images_tensor(n)
    output_col = ['camera_blocked', 'door_open', 'person_present', 'door_locked',
                  'face_x1', 'face_y1', 'face_w', 'face_h',
                  'facial_characteristics',
                  'body_x1', 'body_y1', 'body_w', 'body_h', 'shirt_type']
    type_guesser = TypeGuesser()
    type_guesser.type_mapping['camera_blocked'] = 'binary'
    type_guesser.type_mapping['door_open'] = 'binary'
    type_guesser.type_mapping['person_present'] = 'binary'
    type_guesser.type_mapping['door_locked'] = 'binary'
    type_guesser.type_mapping['face_x1'] = 'continuous'
    type_guesser.type_mapping['face_y1'] = 'continuous'
    type_guesser.type_mapping['face_w'] = 'continuous'
    type_guesser.type_mapping['face_h'] = 'continuous'
    type_guesser.type_mapping['body_x1'] = 'continuous'
    type_guesser.type_mapping['body_y1'] = 'continuous'
    type_guesser.type_mapping['body_w'] = 'continuous'
    type_guesser.type_mapping['body_h'] = 'continuous'
    type_guesser.type_mapping['syn_img'] = 'img'
    type_guesser.type_mapping['shirt_type'] = 'category'
    type_guesser.type_mapping['facial_characteristics'] = 'multilabel'

    values_converter = ValuesConverter()
    values_converter.type_mapping['img'] = lambda x: imgs
    values_converter.type_mapping['binary'] = binary_value_converter
    values_converter.type_mapping['continuous'] = ImageCoordinatesValuesConverter(dim=64)
    values_converter.type_mapping['category'] = classification_converter

    multilabel_converter = MultiLabelValuesConverter()
    values_converter.type_mapping['multilabel'] = multilabel_converter

    task_converter = TaskConverter()

    task_converter.type_mapping['binary'] = To(BinaryClassificationTask,
                                               module_supplier=partial(nn.Linear, in_features=256, out_features=1))
    task_converter.type_mapping['continuous'] = To(BoundedRegressionTask,
                                                   module_supplier=partial(nn.Linear, in_features=256, out_features=1),
                                                   decoder_supplier=partial(BoundedRegressionDecoder, scale=64))
    n_classes = classification_converter(df['shirt_type']).max().item() + 1
    task_converter.type_mapping['category'] = To(ClassificationTask,
                                                 module_supplier=partial(nn.Linear,
                                                                         in_features=256, out_features=n_classes))
    n_classes = multilabel_converter(df['facial_characteristics']).shape[1]
    task_converter.type_mapping['multilabel'] = To(MultilabelClassificationTask,
                                                   module_supplier=partial(nn.Linear,
                                                                           in_features=256, out_features=n_classes))

    converters = Converters()
    converters.task = task_converter
    converters.type = type_guesser
    converters.values = values_converter

    project = Project(df, input_col='syn_img', output_col=output_col, converters=converters,
                      project_dir='./security_project')

    @project.add_flow
    def face_regression(flow, x, out):
        out += flow.face_x1(x.face_localization)
        out += flow.face_y1(x.face_localization)
        out += flow.face_w(x.face_localization)
        out += flow.face_h(x.face_localization)
        out += flow.facial_characteristics(x.features)
        return out

    @project.add_flow
    def body_regression(flow, x, out):
        out += flow.body_x1(x.body_localization)
        out += flow.body_y1(x.body_localization)
        out += flow.body_w(x.body_localization)
        out += flow.body_h(x.body_localization)
        out += flow.shirt_type(x.features)
        return out

    @project.add_flow
    def person_regression(flow, x, out):
        out += flow.face_regression(x)
        out += flow.body_regression(x)
        return out

    @project.add_flow
    def full_flow(flow, x, out):
        out += flow.camera_blocked(x.features)
        out += flow.door_open(x.features) | (~out.camera_blocked)
        out += flow.door_locked(x.features) | (~out.door_open)
        out += flow.person_present(x.features) | out.door_open
        out += flow.person_regression(x) | out.person_present
        return out

    dataset = project.get_full_flow().get_dataset()
    train_dataset, val_dataset = torch_split_dataset(dataset, random_state=42)
    train_loader = DataLoader(train_dataset, batch_size=32 * torch.cuda.device_count(), shuffle=True)
    val_loader = DataLoader(val_dataset, batch_size=32 * torch.cuda.device_count(), shuffle=False)
    nested_loaders = OrderedDict({
        'train': train_loader,
        'valid': val_loader
    })

    class SecurityModule(nn.Module):

        def __init__(self):
            super().__init__()
            self.seq = nn.Sequential(
                nn.Conv2d(3, 64, kernel_size=5),
                nn.ReLU(inplace=True),
                nn.Conv2d(64, 128, kernel_size=5),
                nn.ReLU(inplace=True),
                nn.AvgPool2d(2),
                nn.Conv2d(128, 128, kernel_size=5),
                nn.ReLU(inplace=True),
                nn.Conv2d(128, 256, kernel_size=5),
                nn.AvgPool2d(2),
                nn.ReLU(inplace=True),
            )

            self.features_seq = nn.Sequential(
                nn.Conv2d(256, 256, kernel_size=5),
                nn.ReLU(inplace=True),
                nn.Conv2d(256, 256, kernel_size=5),
                nn.ReLU(inplace=True),
                nn.AdaptiveAvgPool2d(1),
                nn.Flatten(),
            )

            self.face_localization_seq = nn.Sequential(
                nn.Conv2d(256, 256, kernel_size=5),
                nn.ReLU(inplace=True),
                nn.Conv2d(256, 256, kernel_size=5),
                nn.ReLU(inplace=True),
                nn.AdaptiveAvgPool2d(1),
                nn.Flatten(),
            )

            self.body_localization_seq = nn.Sequential(
                nn.Conv2d(256, 256, kernel_size=5),
                nn.ReLU(inplace=True),
                nn.Conv2d(256, 256, kernel_size=5),
                nn.ReLU(inplace=True),
                nn.AdaptiveAvgPool2d(1),
                nn.Flatten(),
            )

            self.flow_module = full_flow.torch()

        def forward(self, x):
            res = {}
            common = self.seq(x['syn_img'])
            res['features'] = self.features_seq(common)
            res['face_localization'] = self.face_localization_seq(common)
            res['body_localization'] = self.body_localization_seq(common)
            res['gt'] = x.get('gt')
            return self.flow_module(res)

    model = SecurityModule()
    datasets = {
        'train': train_dataset,
        'valid': val_dataset,
        'infer': val_dataset
    }
    return model, nested_loaders, datasets, project



from dataclasses import dataclass, field
from typing import Dict, Optional

import torch
from torch import nn

from dnn_cool.dsl import IFeaturesDict, IFlowTaskResult, ICondition
from dnn_cool.utils import to_broadcastable_shape


class SigmoidAndMSELoss(nn.Module):

    def __init__(self, reduction):
        super().__init__()
        self.sigmoid = nn.Sigmoid()
        self.mse = nn.MSELoss(reduction=reduction)

    def forward(self, output, target):
        activated_output = self.sigmoid(output)
        return self.mse(activated_output, target)


class Identity(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x


# noinspection PyBroadException
def find_arg_with_gt(args, is_kwargs):
    new_args = {} if is_kwargs else []
    gt = None
    for arg in args:
        if is_kwargs:
            arg = args[arg]
        try:
            gt = arg['gt']
            new_args.append(arg['value'])
        except:
            new_args.append(arg)
    return gt, new_args


def select_gt(a_gt, k_gt):
    if a_gt is None and k_gt is None:
        return {}

    if a_gt is None:
        return k_gt

    return a_gt


def find_gt_and_process_args_when_training(*args, **kwargs):
    a_gt, args = find_arg_with_gt(args, is_kwargs=False)
    k_gt, kwargs = find_arg_with_gt(kwargs, is_kwargs=True)
    return args, kwargs, select_gt(a_gt, k_gt)


class ModuleDecorator(nn.Module):

    def __init__(self, task, prefix):
        super().__init__()
        self.prefix = prefix
        self.task_name = task.get_name()
        self.module = task.torch()
        self.activation = task.get_activation()
        self.decoder = task.get_decoder()
        self.dropout_mc = task.get_dropout_mc()

    def forward(self, *args, **kwargs):
        key = self.prefix + self.task_name
        logits = self.module(*args, **kwargs)
        condition = OnesCondition(key)
        if self.training:
            return LeafModuleOutput(path=key, logits=logits, precondition=condition,
                                    activated=None, decoded=None, dropout_samples=None)
        activated_logits = self.activation(logits) if self.activation is not None else logits
        decoded_logits = self.decoder(activated_logits) if self.decoder is not None else activated_logits
        condition = OnesCondition(key)
        dropout_samples = None if self.dropout_mc is None else self.dropout_mc.create_samples(self.module,
                                                                                              self.activation,
                                                                                              *args, **kwargs)
        return LeafModuleOutput(path=key, logits=logits, precondition=condition,
                                activated=activated_logits, decoded=decoded_logits, dropout_samples=dropout_samples)


class Condition(ICondition):

    def get_precondition(self, data):
        raise NotImplementedError()

    def to_mask(self, data):
        raise NotImplementedError()

    def __invert__(self):
        return NegatedCondition(self)

    def __and__(self, other):
        return AndCondition(self, other)


@dataclass
class OnesCondition(Condition):
    path: str

    def get_precondition(self, data):
        return OnesCondition(self.path)

    def to_mask(self, data):
        if '_availability' in data:
            return data['_availability'][self.path]
        return torch.ones_like(data[self.path]).bool()


@dataclass()
class NegatedCondition(Condition):
    precondition: Condition

    def get_precondition(self, data):
        return self.precondition.get_precondition(data)

    def to_mask(self, data):
        mask = self.precondition.to_mask(data)
        precondition = self.get_precondition(data).to_mask(data)
        mask[~precondition] = False
        mask[precondition] = ~mask[precondition]
        return mask


@dataclass
class LeafCondition(Condition):
    path: str

    def get_precondition(self, data):
        return OnesCondition(self.path)

    def to_mask(self, data):
        return data[self.path].clone()


@dataclass()
class NestedCondition(Condition):
    path: str
    parent: Condition

    def get_precondition(self, data):
        return self.parent

    def to_mask(self, data):
        mask = data[self.path].clone()
        precondition = self.get_precondition(data).to_mask(data)
        mask[~precondition] = False
        return mask


@dataclass()
class AndCondition(Condition):
    condition_one: Condition
    condition_two: Condition

    def get_precondition(self, data):
        return self.condition_one.get_precondition(data) & self.condition_two.get_precondition(data)

    def to_mask(self, data):
        mask_one = self.condition_one.to_mask(data)
        mask_two = self.condition_two.to_mask(data)
        mask_one, mask_two = to_broadcastable_shape(mask_one, mask_two)
        return mask_one & mask_two


@dataclass
class LeafModuleOutput(IFlowTaskResult):
    path: str
    logits: torch.Tensor
    precondition: Condition
    activated: Optional[torch.Tensor]
    decoded: Optional[torch.Tensor]
    dropout_samples: Optional[torch.Tensor]

    def add_to_composite(self, composite_module_output):
        composite_module_output.logits[self.path] = self.logits
        composite_module_output.activated[self.path] = self.activated
        composite_module_output.decoded[self.path] = self.decoded
        composite_module_output.preconditions[self.path] = self.precondition
        composite_module_output.dropout_samples[self.path] = self.dropout_samples

    def __or__(self, precondition: Condition):
        """
        Note: this has the meaning of a a precondition, not boolean or.
        :param precondition: precondition
        :return: self
        """
        self.precondition &= precondition
        return self


def _copy_to_self(self_arr, other_arr):
    for key, value in self_arr.items():
        assert key not in other_arr, f'The key {key} has been added twice in the same workflow!.'
        other_arr[key] = value


@dataclass
class CompositeModuleOutput(IFlowTaskResult):
    training: bool
    gt: Dict[str, torch.Tensor]
    prefix: str
    logits: Dict[str, torch.Tensor] = field(default_factory=lambda: {})
    activated: Dict[str, torch.Tensor] = field(default_factory=lambda: {})
    decoded: Dict[str, torch.Tensor] = field(default_factory=lambda: {})
    preconditions: Dict[str, Condition] = field(default_factory=lambda: {})
    # A number of predictions for the same tasks, with different parts dropped out
    dropout_samples: Dict[str, torch.Tensor] = field(default_factory=lambda: {})

    def add_to_composite(self, other):
        _copy_to_self(self.logits, other.logits)
        _copy_to_self(self.activated, other.activated)
        _copy_to_self(self.decoded, other.decoded)
        _copy_to_self(self.preconditions, other.preconditions)

    def __iadd__(self, other):
        other.add_to_composite(self)
        return self

    def __getattr__(self, item):
        full_path = self.prefix + item
        parent_precondition = self.preconditions.get(full_path)
        if parent_precondition is None:
            return LeafCondition(path=full_path)
        return NestedCondition(path=full_path, parent=parent_precondition)

    def __or__(self, precondition: Condition):
        """
        Note: this has the meaning of a a precondition, not boolean or.
        :param precondition: precondition tensor
        :return: self
        """
        for key, value in self.preconditions.items():
            current_precondition = self.preconditions.get(key)
            if current_precondition is None:
                self.preconditions[key] = precondition
            else:
                self.preconditions[key] &= precondition
        return self

    def reduce(self):
        if len(self.prefix) != 0:
            # if this is not root, do not reduce yet.
            return self
        inference_without_gt = self.gt is None
        preconditions_source = self.decoded if inference_without_gt else self.gt
        res = self.decoded if inference_without_gt else self.logits

        if inference_without_gt and not self.training:
            for key, value in self.preconditions.items():
                if value is not None:
                    self.preconditions[key] = value.to_mask(preconditions_source)
            return self
        for key, value in self.preconditions.items():
            if value is not None:
                res[f'precondition|{key}'] = value.to_mask(preconditions_source)
        if not inference_without_gt:
            res['gt'] = self.gt
        return res


@dataclass
class FeaturesDict(IFeaturesDict):
    data: Dict

    def __init__(self, data):
        self.data = data

    def __getattr__(self, item):
        return self.data[item]


class TaskFlowModule(nn.Module):

    def __init__(self, task_flow, prefix=''):
        super().__init__()
        self._task_flow = task_flow
        # Save a reference to the flow function of the original class
        # We wi.ll then call it by replacing the self, this way effectively running
        # it with this class. And this class stores Pytorch modules as class attributes
        self.flow = task_flow.get_flow_func()
        self.prefix = prefix

        for key, task in task_flow.tasks.items():
            if not task.has_children():
                instance = ModuleDecorator(task, prefix)
            else:
                instance = TaskFlowModule(task, prefix=f'{prefix}{task.get_name()}.')
            setattr(self, key, instance)

    def forward(self, x):
        if isinstance(x, FeaturesDict):
            x = x.data

        out = CompositeModuleOutput(training=self.training, gt=x.get('gt'), prefix=self.prefix)
        composite_module_output = self.flow(self, FeaturesDict(x), out)
        outputs = composite_module_output.reduce()
        return outputs

    def load_tuned(self, tuned_params):
        decoders = self._get_all_decoders()
        for key, decoder in decoders.items():
            decoder.load_tuned(tuned_params[key])

    def _get_all_decoders(self):
        decoders = {}
        for task_name, task in self._task_flow.tasks.items():
            if task.has_children():
                attr: TaskFlowModule = getattr(self, task_name)
                decoders.update(attr._get_all_decoders())
            else:
                decoders[self.prefix + task_name] = task.get_decoder()
        return decoders



from setuptools import setup, find_packages
import pathlib
import dnn_cool

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="dnn_cool",
    version=dnn_cool.__version__,
    description="DNN.Cool: Multi-task learning for Deep Neural Networks (DNN).",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/hristo-vrigazov/dnn.cool",
    author="Hristo Vrigazov",
    author_email="hvrigazov@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent"
    ],
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=["torch",
                      "catalyst == 20.12",
                      "pandas",
                      "numpy",
                      "tqdm",
                      "scikit_learn",
                      "treelib",
                      "matplotlib",
                      "joblib"],
    extras_require={
        'nlp': ['transformers'],
        'interpretability': ['captum']
    }
)
